

set -ex



pip check
export OPENBLAS_NUM_THREADS=1
export OMP_NUM_THREADS=1
export MPLBACKEND=Agg
SKIMAGE_TEST_STRICT_WARNINGS=0 pytest --verbose --pyargs skimage -k "not (_not_a_real_test or test_apply_parallel_rgb or test_cut_normalized or test_ncut_stable_subgraph or test_reproducibility or test_analytical_moments_calculation)"
exit 0
